
library(shiny)
library(data.table)
library(ggplot2)

# app code prior to clean up

# install.packages("bit64")  # is this needed?

options(shiny.maxRequestSize = 10000000000000*1024^2)

# models options for fitting the components
modChoices <- c("none" = "none", "Linear" = "lin", "Exp" = "exp1", "Bi-Exp" = "biExp", "Plateau Plus Bi-Exp" = "platPlus")

readtextdata <- function(fileName) {
  
  fH <- file(fileName, "r")
  lineNum <- 0
  while ( TRUE ) {
    line <- readLines(fH, n=1)
    
    if ( line == ">>>>>Begin Spectral Data<<<<<") {
      break
    }
    lineNum <- lineNum  + 1
  }
  close(fH)
  RawData <- fread(fileName, skip=lineNum, header=FALSE)
  wavelengths <- as.numeric(RawData[1,3:ncol(RawData)])
  Data <- t(as.matrix(RawData[2:nrow(RawData), 3:ncol(RawData)]))
  return(list(wave=wavelengths,Mixed=Data))
}

readcsvdata <- function(fileName) {
  RawData <- fread(fileName, header = FALSE)
  wavelengths <- as.numeric(RawData[1,2:ncol(RawData)])
  Data <- t(as.matrix(RawData[2:nrow(RawData), 2:ncol(RawData)]))
  return(list(wave=wavelengths,Mixed=Data))
}

getPred <- function(modType,tVals,coefVals) {
  D <- data.table(tPoint=tVals, y=coefVals)
  
  if (modType == "lin") {
    # linear
    fit1 <- lm(data=D, y ~ tPoint)
  } else if (modType == "exp1") {
    # one exponential
    fit1 <- nls(data = D, y ~ (y0 - plat)*exp(-k*tPoint) + plat, start = list(plat=475, y0=525, k=0.004))
  } else if (modType == "biExp") {
    # two exponential
    fit1 <- nls(data=D, y ~ SSbiexp(tPoint, A1, r1,A2, r2))
  } else if (modType == "platPlus") {
    # plateau plus two exponential
    fit1 <- nls(data = D, y ~ plat + A*exp(-kFast*tPoint) + B*exp(-kSlow*tPoint), start = list(A=100, B=5, plat=480, kFast=0.02, kSlow=0.001))
  } else {
    fit1 <- NULL
    # error since I can make sure there are only these inputs
  }
  
  if (is.null(fit1)) {
    myOut <- rep(NaN, length(tVals))
  } else {
  myOut <- predict(fit1)
  }
}
  

ui <- navbarPage("Linear Spectral Unmixing",
                 
                 tabPanel("Data Info",
                          sidebarPanel(
                            # Data file as TXT or CSV
                            fileInput('specFile', 'Choose Data file (TXT or CSV)'),
                            
                            # Reference file
                            fileInput('refFile', 'Choose CSV reference file'),
                            
                            strong(helpText("What is the range of wavelength you would like to analyze?")),
                            textInput("lowerWave", "Enter the lower limit:", 0),
                            textInput("upperWave", "Enter the upper limit:", 1500),
                            
                            hr(),
                            
                            
                            conditionalPanel(
                              condition = "output.nVar > '1'",
                              strong(helpText("Reference labels may be changed")),
                              textInput("name1", "enter a new name")
                            ),
                            conditionalPanel(
                              condition = "output.nVar > '2'",
                              textInput("name2", "enter a new name")
                            ),
                            conditionalPanel(
                              condition = "output.nVar > '3'",
                              textInput("name3", "enter a new name")
                            ),
                            conditionalPanel(
                              condition = "output.nVar > '4'",
                              textInput("name4", "enter a new name")
                            )
                            
                            
                          ),
                          
                          
                          mainPanel(
                          
                            tableOutput("dataFileInfo"),
                            tableOutput("DataCheck"),
                            hr(),
                            tableOutput("refFileInfo"),
                            textOutput("nVar"),
                            tableOutput("RefCheck")
                          ),
                 ),
                 
                 tabPanel(
                   "Linear Out",
                   sidebarPanel(
                     #Use Numeric Input
                     numericInput("time_point", "Enter a time point:", min = 0, value = 0),
                     
                   ),
                   mainPanel(
                     textOutput("testing"),
                     plotOutput("Regression_Line"),
                     tableOutput("waveStuff"),
                     
                   )
                 ), 
                 
                 tabPanel(
                   "Coeff Plots",
                   sidebarPanel(
                     conditionalPanel(
                       condition = "output.nVar > '1'",
                       selectInput("pick1","Pick model to fit to component 1", modChoices)
                     ),

                     conditionalPanel(
                       condition = "output.nVar > '2'",
                       selectInput("pick2","Pick model to fit to component 2", modChoices)
                     ),
                     
                     conditionalPanel(
                       condition = "output.nVar > '3'",
                       selectInput("pick3","Pick model to fit to component 3", modChoices)
                     ),
                     
                     conditionalPanel(
                       condition = "output.nVar > '4'",
                       selectInput("pick4","Pick model to fit to component 4", modChoices)
                     ),


                   ),
                   mainPanel(
                   plotOutput("coeffPlot"),
                   tableOutput("fitAllCheck"),
                   
                   )
                 ),
                 
                 tabPanel(
                   "Output table",
                   mainPanel(
                     textInput("outFileName", "What would you like to call the output file?", "Output"),
                     downloadButton("downloadData","Download dataset."),
                     tags$hr(),
                     tableOutput("outTabCheck"),
                   )
                 )
                 
)

server <- function(input, output, session){
  
  FitLines <- reactive({
    if (nChan() == 2) {
      D <- data.table(fit1values(), fit2values())
    } else if (nChan() == 3 ) {
      D <- data.table(fit1values(), fit2values(), fit3values())
    } else if (nChan() == 4 ) {
      D <- data.table(fit1values(), fit2values(), fit3values(), fit4values())
    } else {
      D <- data.table()
    }
    
    names(D) <- refNames()[1:nChan()]
    D[, t := tPoints()]
    myOut <- D
  }) 
  
  output$fitAllCheck <- renderTable(FitLines()[1:6,])
  
  tPoints <- reactive({
    1:dim(fitAll()$coefficients)[2]
  })
  
  
  fit1values <- reactive({
    pred <- getPred(input$pick1, tVals = tPoints(), coefVals = fitAll()$coefficients[2,]) 
  }
  )
  
  fit2values <- reactive({
    pred <- getPred(input$pick2, tVals = tPoints(), coefVals = fitAll()$coefficients[3,]) 
  }
  )
  
  fit3values <- reactive({
    pred <- getPred(input$pick3, tVals = tPoints(), coefVals = fitAll()$coefficients[4,]) 
  }
  )
  
  fit4values <- reactive({
    pred <- getPred(input$pick4, tVals = tPoints(), coefVals = fitAll()$coefficients[5,]) 
  }
  )
  
  #Downloadable file
  output$downloadData <- downloadHandler(
    filename = function(){
      paste(input$outFileName, "csv", sep = ".")
    },
    content = function(file){
      dfOut <- outTab()
      write.table(dfOut, file, sep = ",", row.names = FALSE)
    }
  )
  
  output$outTabCheck <- renderTable(outTab()[1:15,])
  
  outTab <- reactive({
    modCoef <- data.table(t(fitAll()$coefficients))
    
    # drop "channels" in the names that come fromm lm fit
    setnames(modCoef, names(modCoef), gsub("channels", "", names(modCoef)))
    
    # change originals to new reference names from the user
    setnames(modCoef, origNames()[-1], refNames())
    modCoef[, t := tPoints()]
    
    upIndx <- waveInfo()$upIndx 
    loIndx <- waveInfo()$loIndx 
    
    # calculate R sqr
    nWave <- length(SpecData()$wave[loIndx:upIndx])
    md <- SpecData()$Mixed #md <- t(SpecData()$Mixed) #md <- t(as.matrix(SpecData()$Mixed))
    cm <- colMeans(md)
    cm <- matrix(rep(cm,each=nWave),nrow=nWave)
    R <- fitAll()$residuals
    RsqAll <- 1-colSums(R^2)/colSums((cm-md[loIndx:upIndx,])^2)
    modCoef[, rsqVals := RsqAll]  
    
    predVals <- FitLines()
    names(predVals) <- paste0(names(predVals),'_pred')
    
    modCoef <- cbind(modCoef, predVals)
    
    
    for (rN in refNames()) {
      mycols <- c(rN, paste0(rN,"_pred"))
      Temp <- modCoef[, ..mycols]
      setnames(Temp, names(Temp), c("x", "pred"))
      percName <- paste0(rN,"percC")
      Temp[, percC := (x-pred)/pred*100]
      modCoef[, paste(percName) := Temp[, percC]]
    }

    modCoef[, t_pred := NULL]
    
  })
  
  output$coeffPlot <- renderPlot({
    temp <- data.table(t(fitAll()$coefficients))
    temp[, '(Intercept)' := NULL]
    names(temp) <- refNames()[1:nChan()] #gsub("channels", "", names(temp))
    temp[, t := tPoints()]
    temp <- melt(temp, id.vars = "t", variable.name = "channel", value.name = "Coeff")
    
    tempFit <- melt(FitLines(), id.vars = "t", variable.name = "channel", value.name = "Coeff")
    
    ggplot(data=temp, aes(x=t, y=Coeff)) + geom_line() + 
      facet_wrap(~channel, scales = "free") + theme(strip.text = element_text(size=20)) +
      geom_line(data=tempFit[is.nan(Coeff) != TRUE,], size=2, color="green")
    
  })
  
  nChan <- reactive(
    length(names(RefData()))-1
  )
  
  output$testing <- renderText("Model fits (red) vs. Data (black) at each time point")
  
  output$Regression_Line <- renderPlot({
    
    plot(x=SpecData()$wave, y=SpecData()$Mixed[, time()])
    lines(x=SpecData()$wave[waveInfo()$loIndx:waveInfo()$upIndx], y=fitAll()$fitted.values[,time()], col = "red")
    
  })  
  
  time <- reactive({
    as.numeric(input$time_point) +1
  })
  
  # keep the min and max values and indices in one variable
  waveInfo <- reactive({
    
    minWaveData <- min(SpecData()$wave)
    maxWaveData <- max(SpecData()$wave)
    
    minWaveUser <- as.numeric(input$lowerWave)
    maxWaveUser <- as.numeric(input$upperWave)
    
    upIndx <- tail(which(SpecData()$wave <= maxWaveUser),1 )
    loIndx <- head(which(SpecData()$wave >= minWaveUser),1 )
    
    return(list(loIndx = loIndx, upIndx = upIndx, minWaveData = minWaveData, maxWaveData = maxWaveData))
  })
  
  output$waveStuff <- renderTable(waveInfo())
  
  # check is regression will run
  fitAll <- reactive({
    
    MD <- SpecData()$Mixed
    RD <- RefData()
    
    RD <- RD[`Wavelength (nm)` >= waveInfo()$minWaveData & `Wavelength (nm)` <= waveInfo()$maxWaveData,]
    
    upIndx <- waveInfo()$upIndx 
    loIndx <- waveInfo()$loIndx 
    
    
   
    channels <- as.matrix(RD[loIndx:upIndx, -1])
    
    
    if (dim(channels)[1] == dim(MD[loIndx:upIndx,])[1]) {
      #fitLM1 <- lm(MD ~ channels)
      fitLM3 <- lm(MD[loIndx:upIndx,] ~ channels)
    } else {
      #fitLM1 <- NULL
      fitLM3 <- NULL
    }
    
    fitAll <- fitLM3
  })
  
 
  
  # Read Spectral Data to analyze.  Returns data and wavelengths
  SpecData <- reactive({
    dataFile <- input$specFile
    if(is.null(dataFile)){return()}
    if (dataFile$type == "text/plain" ) {
      Raw <- readtextdata(input$specFile$datapath)
    } else {
      Raw <- readcsvdata(input$specFile$datapath)
    }
  })
  
  # output checks of Spectral Data
  output$DataCheck <- renderTable(SpecData()$Mixed[1:3,1:5])
  output$dataFileInfo <- renderTable(input$specFile)
  
  
  # Read in Reference File
  RefData <- reactive({
    refFile <- input$refFile
    if(is.null(refFile)){return()}
    TempRD <- fread(file = refFile$datapath)
    setnames(TempRD, names(TempRD), gsub("-", "_", names(TempRD)))
  })
  
  # output check of Reference file
  output$refFileInfo <- renderTable(input$refFile)
  output$RefCheck <- renderTable(RefData()[1:5,])
  output$nVar <- renderText(length(names(RefData())))
  
  # names of the colors in the refenece file
  origNames <- reactive(names(RefData()))
  
  # new names
  refNames <- reactive({
    rNames <- origNames()[-1]
    if (nchar(input$name1) > 0) {rNames[1] <- input$name1}
    if (nchar(input$name2) > 0) {rNames[2] <- input$name2}
    if (nchar(input$name3) > 0) {rNames[3] <- input$name3}
    if (nchar(input$name4) > 0) {rNames[4] <- input$name4}
    rNames
  })
  
  # rename refernce names if desired
  observe({
    updateTextInput(session, "name1", label = paste("New label for", origNames()[2]))
    updateTextInput(session, "name2", label = paste("New label for", origNames()[3]))
    updateTextInput(session, "name3", label = paste("New label for", origNames()[4]))
    updateTextInput(session, "name4", label = paste("New label for", origNames()[5]))
    updateSelectInput(session, "pick1", label = paste("Pick model to fit to ", refNames()[1]))
    updateSelectInput(session, "pick2", label = paste("Pick model to fit to ", refNames()[2]))
    updateSelectInput(session, "pick3", label = paste("Pick model to fit to ", refNames()[3]))
    updateSelectInput(session, "pick4", label = paste("Pick model to fit to ", refNames()[4]))
  })
  
  
}

shinyApp(ui,server)

